/*!
 * Bootstrap v1.4.0
 *
 * Copyright 2011 Twitter, Inc
 * Licensed under the Apache License v2.0
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Heavily modified by Kyle Barbary for the AstroPy Project for use with Sphinx.
 *
 * Slightly modified by Marshall Perrin for use with WebbPSF & POPPY with a larger
 * font size and some other display tweaks
 */

@import url("basic.css");

body {
  background-color: #ffffff;
  margin: 0;
  font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
  font-size: 15px;
  font-weight: normal;
  line-height: 20px;
  color: #404040;
}

/* Hyperlinks ----------------------------------------------------------------*/

a {
  color: #0069d6;
  text-decoration: none;
  line-height: inherit;
  font-weight: inherit;
}

a:hover {
  color: #00438a;
  text-decoration: underline;
}

/* Typography ----------------------------------------------------------------*/

h1,h2,h3,h4,h5,h6 {
  color: #404040;
  margin: 0.7em 0 0 0;
  line-height: 1.5em;
}
h1 {
  font-size: 24px;
  margin: 0;
}
h2 {
  font-size: 21px;
  line-height: 1.2em;
  margin: 1em 0 0.5em 0;
  border-bottom: 1px solid #404040;
}
h3 {
  font-size: 18px;
}
h4 {
  font-size: 16px;
}
h5 {
  font-size: 14px;
}
h6 {
  font-size: 13px;
  text-transform: uppercase;
}

p {
  font-size: 15px;
  font-weight: normal;
  line-height: 20px;
  margin-top: 0px;
  margin-bottom: 9px;
}

ul, ol {
  margin-left: 0;
  padding: 0 0 0 25px;
}
ul ul, ul ol, ol ol, ol ul {
  margin-bottom: 0;
}
ul {
  list-style: disc;
}
ol {
  list-style: decimal;
}
li {
  line-height: 18px;
  color: #404040;
}
ul.unstyled {
  list-style: none;
  margin-left: 0;
}
dl {
  margin-bottom: 18px;
}
dl dt, dl dd {
  line-height: 18px;
}
dl dd {
  margin-left: 9px;
}
hr {
  margin: 20px 0 19px;
  border: 0;
  border-bottom: 1px solid #eee;
}
strong {
  font-style: inherit;
  font-weight: bold;
}
em {
  font-style: italic;
  font-weight: inherit;
  line-height: inherit;
}
.muted {
  color: #bfbfbf;
}

address {
  display: block;
  line-height: 18px;
  margin-bottom: 18px;
}
code, pre {
  padding: 0 3px 2px;
  font-family: Monaco, Andale Mono, Courier New, monospace;
  font-size: 12px;
  -webkit-border-radius: 3px;
  -moz-border-radius: 3px;
  border-radius: 3px;
}
code {
  color: rgba(0, 0, 0, 0.75);
  padding: 1px 3px;
}
pre {
  display: block;
  padding: 8.5px;
  margin: 0 0 18px;
  line-height: 18px;
  font-size: 12px;
  border: 1px solid #ddd;
  border: 1px solid rgba(0, 0, 0, 0.12);
  -webkit-border-radius: 3px;
  -moz-border-radius: 3px;
  border-radius: 3px;
  white-space: pre;
  white-space: pre-wrap;
  word-wrap: break-word;
}

img {
  margin: 9px 0;
}

/* format inline code with a rounded box */
tt {
    margin: 0 2px;
    padding: 0 5px;
    border: 1px solid #ddd;
    border: 1px solid rgba(0, 0, 0, 0.12);
    border-radius: 3px;
    font-size: 0.95em;
}

/* all code has same box background color, even in headers */
h1 tt, h2 tt, h3 tt, h4 tt, h5 tt, h6 tt, pre, code, tt {
  background-color: #f8f8f8;
}

/* override box for links & other sphinx-specifc stuff */
tt.xref, a tt, tt.descname, tt.descclassname {
  padding: 0 1px 0 1px;
  border: none;
}

/* override box for related bar at the top of the page */
.related tt {
  border: none;
  padding: 0 1px 0 1px;
  background-color: transparent;
  font-weight: bold;
}

th {
    background-color: #dddddd;
}

.viewcode-back {
    font-family: sans-serif;
}

div.viewcode-block:target {
    background-color: #f4debf;
    border-top: 1px solid #ac9;
    border-bottom: 1px solid #ac9;
}

table.docutils {
    border-spacing: 5px;
    border-collapse: separate;
}

/* Topbar --------------------------------------------------------------------*/

div.topbar {
  height: 40px;
  position: absolute;
  top: 0;
  left: 0;
  right: 0;
  z-index: 10000;
  padding: 0px 10px;
  background-color: #222;
  background-color: #222222;
  background-repeat: repeat-x;
  background-image: -khtml-gradient(linear, left top, left bottom, from(#333333), to(#222222));
  background-image: -moz-linear-gradient(top, #333333, #222222);
  background-image: -ms-linear-gradient(top, #333333, #222222);
  background-image: -webkit-gradient(linear, left top, left bottom, color-stop(0%, #333333), color-stop(100%, #222222));
  background-image: -webkit-linear-gradient(top, #333333, #222222);
  background-image: -o-linear-gradient(top, #333333, #222222);
  background-image: linear-gradient(top, #333333, #222222);
  filter: progid:DXImageTransform.Microsoft.gradient(startColorstr='#333333', endColorstr='#222222', GradientType=0);
}

div.topbar a.brand {
  font-family: 'Source Sans Pro', sans-serif;
  font-size: 26px;
  color: #ffffff;
  font-weight: 600;
  text-decoration: none;
  float: left;
  display: block;
  height: 32px;
  padding: 8px 12px 0px 45px;
  margin-left: -10px;
  background: transparent url("webbpsf_logo_32.png") no-repeat 10px 4px;
}

#logotext1 {
}

#logotext2 {
    font-weight:200;
    color: #ff5000;
}
#logotext3 {
    font-weight:200;
}

div.topbar .brand:hover, div.topbar ul li a.homelink:hover {
  background-color: #333;
  background-color: rgba(255, 255, 255, 0.05);
}

div.topbar ul {
    font-size: 110%;
    list-style: none;
    margin: 0;
    padding: 0 0 0 10px;
    float: right;
    color: #bfbfbf;
    text-align: center;
    text-decoration: none;
    height: 100%;
}
div.topbar ul li {
    float: left;
    display: inline;
    height: 30px;
    margin: 5px;
    padding: 0px;
}

div.topbar ul li a {
  color: #bfbfbf;
  text-decoration: none;
  padding: 5px;
  display: block;
  height: auto;
  text-align: center;
  vertical-align: middle;
  border-radius: 4px;
}

div.topbar ul li a:hover {
  color: #ffffff;
  text-decoration: none;
}

div.topbar ul li a.homelink {
  width: 112px;
  display: block;
  height: 20px;
  padding: 5px 0px;
  /* background: transparent url("astropy_linkout_20.png") no-repeat 10px 5px; */
}

div.topbar form {
  text-align: left;
  margin: 0 0 0 5px;
  position: relative;
  filter: alpha(opacity=100);
  -khtml-opacity: 1;
  -moz-opacity: 1;
  opacity: 1;
}

div.topbar input {
  background-color: #444;
  background-color: rgba(255, 255, 255, 0.3);
  font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
  font-size: normal;
  font-weight: 13px;
  line-height: 1;
  padding: 4px 9px;
  color: #ffffff;
  color: rgba(255, 255, 255, 0.75);
  border: 1px solid #111;
  -webkit-border-radius: 4px;
  -moz-border-radius: 4px;
  border-radius: 4px;
  -webkit-box-shadow: inset 0 1px 2px rgba(0, 0, 0, 0.1), 0 1px 0px rgba(255, 255, 255, 0.25);
  -moz-box-shadow: inset 0 1px 2px rgba(0, 0, 0, 0.1), 0 1px 0px rgba(255, 255, 255, 0.25);
  box-shadow: inset 0 1px 2px rgba(0, 0, 0, 0.1), 0 1px 0px rgba(255, 255, 255, 0.25);
  -webkit-transition: none;
  -moz-transition: none;
  -ms-transition: none;
  -o-transition: none;
  transition: none;
}
div.topbar input:-moz-placeholder {
  color: #e6e6e6;
}
div.topbar input::-webkit-input-placeholder {
  color: #e6e6e6;
}
div.topbar input:hover {
  background-color: #bfbfbf;
  background-color: rgba(255, 255, 255, 0.5);
  color: #ffffff;
}
div.topbar input:focus, div.topbar input.focused {
  outline: 0;
  background-color: #ffffff;
  color: #404040;
  text-shadow: 0 1px 0 #ffffff;
  border: 0;
  padding: 5px 10px;
  -webkit-box-shadow: 0 0 3px rgba(0, 0, 0, 0.15);
  -moz-box-shadow: 0 0 3px rgba(0, 0, 0, 0.15);
  box-shadow: 0 0 3px rgba(0, 0, 0, 0.15);
}


/* Relation bar (breadcrumbs, prev, next) ------------------------------------*/

div.related {
    height: 21px;
    width: auto;
    margin: 0 10px;
    position: absolute;
    top: 42px;
    clear: both;
    left: 0;
    right: 0;
    z-index: 10000;
    font-size: 100%;
    vertical-align: middle;
    background-color: #fff;
    border-bottom: 1px solid #bbb;
}
div.related ul {
    padding: 0;
    margin: 0;
}


/* Footer --------------------------------------------------------------------*/

footer {
  display: block;
  margin: 10px 10px 0px;
  padding: 10px 0 0 0;
  border-top: 1px solid #bbb;
}
.pull-right {
  float: right;
  width: 30em;
  text-align: right;
}


/* Sphinx sidebar ------------------------------------------------------------*/

div.sphinxsidebar {
    font-size: inherit;
    border-radius: 3px;
    background-color: #eee;
    border: 1px solid #bbb;
}

div.sphinxsidebarwrapper {
    padding: 0px 0px 0px 5px;
}

div.sphinxsidebar h3 {
    font-family: 'Trebuchet MS', sans-serif;
    font-size: 1.4em;
    font-weight: normal;
    margin: 5px 0px 0px 5px;
    padding: 0;
    line-height: 1.6em;
}
div.sphinxsidebar h4 {
    font-family: 'Trebuchet MS', sans-serif;
    font-size: 1.3em;
    font-weight: normal;
    margin: 5px 0 0 0;
    padding: 0;
}
div.sphinxsidebar p {
}
div.sphinxsidebar p.topless {
    margin: 5px 10px 10px 10px;
}
div.sphinxsidebar ul {
    margin: 0px 0px 0px 5px;
    padding: 0;
}

div.sphinxsidebar ul ul {
    margin-left: 15px;
    list-style-type: disc;
}

/* If showing the global TOC (toctree),
   color the current page differently */
div.sphinxsidebar a.current {
    color: #404040;
}
div.sphinxsidebar a.current:hover {
    color: #404040;
}


/* document, documentwrapper, body, bodywrapper ----------------------------- */

div.document {
    margin-top: 72px;
    margin-left: 10px;
    margin-right: 10px;
}

div.documentwrapper {
    float: left;
    width: 100%;
}

div.body {
    background-color: #ffffff;
    padding: 0 0 0px 20px;
}

div.bodywrapper {
    margin: 0 0 0 230px;
    max-width: 55em;
}


/* Header links ------------------------------------------------------------- */

a.headerlink {
    font-size: 0.8em;
    padding: 0 4px 0 4px;
    text-decoration: none;
}

a.headerlink:hover {
    background-color: #0069d6;
    color: white;
    text-docoration: none;
}


/* Admonitions and warnings ------------------------------------------------- */

/* Shared by admonitions and warnings */
div.admonition,
div.warning {
    padding: 0px;
    border-radius: 3px;
    -moz-border-radius: 3px;
    -webkit-border-radius: 3px;
}
div.admonition p,
div.warning p {
    margin: 0.5em 1em 0.5em 1em;
    padding: 0;
}
div.admonition pre,
div.warning pre {
    margin: 0.4em 1em 0.4em 1em;
}
div.admonition p.admonition-title,
div.warning p.admonition-title {
    margin: 0;
    padding: 0.1em 0 0.1em 0.5em;
    color: white;
    font-weight: bold;
    font-size: 1.1em;
}
div.admonition ul, div.admonition ol,
div.warning ul, div.warning ol {
    margin: 0.1em 0.5em 0.5em 3em;
    padding: 0;
}

/* Admonitions only */
div.admonition {
    border: 1px solid #609060;
    background-color: #e9ffe9;
}
div.admonition p.admonition-title {
    background-color: #70A070;
}

/* Warnings only */
div.warning {
    border: 1px solid #900000;
    background-color: #ffe9e9;
}
div.warning p.admonition-title {
    background-color: #b04040;
}


/* Figures ------------------------------------------------------------------ */

.figure.align-center {
    clear: none;
}

/* This is a div for containing multiple figures side-by-side, for use with
 * .. container:: figures */
div.figures {
    border: 1px solid #CCCCCC;
    background-color: #F8F8F8;
    margin: 1em;
    text-align: center;
}

div.figures .figure {
    clear: none;
    float: none;
    display: inline-block;
    border: none;
    margin-left: 0.5em;
    margin-right: 0.5em;
}

.field-list th {
    white-space: nowrap;
}
