# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This packages contains affiliated package tests.
"""
# Test code for POPPY - Use the function test() in poppy itself to invoke this.
#
# Test patterns based on http://ssb.stsci.edu/testing/pandokia/docs_new/helper_testfn.html
# and astropy affiliates framework

